@@ -1,11 +1,13 @@
 from synergy_file_reader.tools import (
-		split_well_name, extract_channel,
+		split_well_name, extract_channel, is_sample_id,
 		parse_number, parse_time, parse_timestamp,
-		row_iter,
+		row_iter, VALID_ROWS,
 		LineBuffer,
+		wrap_variant
 	)
 import numpy as np
 from datetime import datetime
+from warnings import warn
 
 format_parsers = [parse_time, parse_number]
 
@@ -16,6 +18,8 @@ timescales = {
 		"days"   : 24*60*60,
 	}
 
+IGNORED_CHANNELS = { "Count", "Mean", "Std Dev", "CV (%)" }
+
 class FormatMismatch(Exception): pass
 class RepeatingData(Exception): pass
 
@@ -71,93 +75,131 @@ class TryFormats(object):
 		else:
 			raise FormatMismatch
 
-class SynergyResult(object):
-	"""
-	A single well- and channel-wise result.
-	
-	You can index this in different ways, using the well F7 for the channel `"OD"` as an example:
-	
-	* Row letter, column number, and channel separately: `result["F",7,"OD"]`
-	* Well identifier in one string: `result["F7","OD"]`
-	* Using lowercase letters to identify the row: `result["f7","OD"]`
-	* If there is only one channel, you do not need to specify it: `result["F7"]`
-	
-	It comes with the following attributes and methods:
-	
-	* `rows` and `cols` are lists of the row letters and column numbers, respectively.
-	* `channels` is a list of all channels for which recordings exists.
-	* `keys` and `values` are methods similar to those for dictionaries, returning iterables of all keys and values respectively.
-
-	"""
-	def __init__(self):
+class SynergyIndexable(object):
+	def __init__(self,sample_ids=None):
 		self.data = {}
 		self.rows = []
 		self.cols = []
-		self.channels = []
+		self.sample_ids = sample_ids
 	
 	def _add_row(self,row):
 		if row not in self.rows:
-			if self.rows:
-				assert len(row)>len(self.rows[-1]) or row>self.rows[-1]
 			self.rows.append(row)
 	
 	def _add_col(self,col):
 		if col not in self.cols:
-			if self.cols:
-				assert col > self.cols[-1]
 			self.cols.append(col)
 	
-	def _add_channel(self,channel):
-		if channel not in self.channels:
-			self.channels.append(channel)
-	
 	def keys(self):
 		return self.data.keys()
 	
 	def values(self):
 		return self.data.values()
 	
-	def _convert_index(self,index):
+	def _normalise_well_index(self,index):
+		"""
+		Returns an iterable of normalised indices from the given index.
+		A normalised index consists of a row, a col, and the further elements of `index`.
+		"""
+		
 		# Avoid string as single index being interpreted as iterable:
-		if isinstance(index,str):
-			index = [index]
+		index = [index] if isinstance(index,str) else list(index)
+		
+		if (
+				    len(index)>=2
+				and (index[0] in VALID_ROWS)
+				and isinstance(index[1],int)
+			):
+				row,col = index[:2]
+				row = row.upper()
+				yield (row,col,*index[2:])
+		elif is_sample_id(index[0]):
+			if self.sample_ids is None:
+				raise ValueError("No layout information available.")
+			results = self.sample_ids[index[0]]
+			for result in results:
+				yield (*result,*index[1:])
 		else:
-			index = list(index)
-		
-		try:
-			row,col = split_well_name(index[0])
-			del index[0]
-		except ValueError:
-			row,col = index[:2]
-			del index[:2]
-		
-		row = row.upper()
-		
-		if len(index)==0:
-			if len(self.channels)==1:
-				channel = self.channels[0]
+			try:
+				row,col = split_well_name(index[0].upper())
+			except ValueError:
+				raise ValueError("Not a valid index.")
 			else:
-				raise ValueError("You must specify a channel as there is more than one in this read.")
-		elif len(index)==1:
-			channel = index[0]
-		else:
-			raise ValueError("Too many indices.")
-		
-		return row,col,channel
-
+				yield (row,col,*index[1:])
+	
+	def _convert_indices(self,index):
+		for row,col,*residual in self._normalise_well_index(index):
+			if residual:
+				raise ValueError("Too many indices.")
+			yield row,col
+	
 	def __setitem__(self,index,result):
-		row,col,channel = self._convert_index(index)
-		if (row,col,channel) in self.keys():
+		indices = list(self._convert_indices(index))
+		if len(indices) != 1:
+			raise ValueError("Index is not unique")
+		if indices[0] in self.keys():
 			raise RepeatingData
-		else:
-			self._add_row(row)
-			self._add_col(col)
-			self._add_channel(channel)
-			self.data[row,col,channel] = result
+		
+		row,col,*residual = indices[0]
+		self._add_row(row)
+		self._add_col(col)
+		self.data[(row,col,*residual)] = result
 	
 	def __getitem__(self,index):
-		return self.data[self._convert_index(index)]
+		indices = list(self._convert_indices(index))
+		if len(indices)==1:
+			return self.data[indices[0]]
+		else:
+			return [ self.data[index] for index in indices ]
 
+class SynergyResult(SynergyIndexable):
+	"""
+	A single well- and channel-wise result.
+	
+	You can index this in different ways, using the well F7 for the channel `"OD"` as an example:
+	
+	* Row letter, column number, and channel separately: `result["F",7,"OD"]`
+	* Well identifier in one string: `result["F7","OD"]`
+	* If there is only one channel, you do not need to specify it: `result["F7"]`
+	* If your plate contains information on sample IDs, you can also use those for indexing: `result["SPL55","OD"]`. If the sample ID is not unique, a list of the respective content of all matching wells will be returned.
+	
+	It comes with the following attributes and methods:
+	
+	* `rows` and `cols` are lists of the row letters and column numbers, respectively.
+	* `channels` is a list of all channels for which recordings exists.
+	* `keys` and `values` are methods similar to those for dictionaries, returning iterables of all keys and values respectively.
+	* If your plate contains information on sample IDs, `sample_ids` contains the mappin of sample IDs to wells.
+	"""
+	
+	def __init__(self,sample_ids=None):
+		super().__init__(sample_ids)
+		self.channels = []
+	
+	def _add_channel(self,channel):
+		if channel not in self.channels:
+			self.channels.append(channel)
+	
+	def _convert_indices(self,index):
+		for row,col,*residual in self._normalise_well_index(index):
+			if len(residual)==0:
+				if len(self.channels)==1:
+					channel = self.channels[0]
+				else:
+					raise ValueError("You must specify a channel as there is more than one in this read.")
+			elif len(residual)==1:
+				channel = residual[0]
+			else:
+				raise ValueError("Too many indices.")
+			
+			yield row,col,channel
+	
+	def __setitem__(self,index,result):
+		if index[-1] not in IGNORED_CHANNELS:
+			super().__setitem__(index,result)
+			_,_,channel = next(self._convert_indices(index))
+			self._add_channel(channel)
+		else:
+			warn(f"Columns/rows, etc. headed with any of {IGNORED_CHANNELS} will be ignored and not parsed.")
 
 class SynergyPlate(SynergyResult):
 	"""
@@ -185,6 +227,8 @@ class SynergyPlate(SynergyResult):
 		self.temperatures = {}
 		self.results = {}
 		self.gains = {}
+		self.sample_ids = None
+		self.layout = None
 	
 	def _add_time(self,time,channel):
 		if self.times is None:
@@ -222,18 +266,21 @@ class SynergyPlate(SynergyResult):
 			else:
 				self[row,col,channel] = np.array([value])
 		else:
-			assert self.times is None
 			self[row,col,channel] = value
 	
 	def _add_result(self,name,row,col,value):
+		if isinstance(name,tuple):
+			self._add_raw_result(name,row,col,value)
+			return
+		
 		try:
-			key,channel = extract_channel(name)
+			key,channel = extract_channel(name,self.channels)
 		except ValueError:
 			self._add_raw_result(name,row,col,value)
 			return
 		
 		if key not in self.results:
-			self.results[key] = SynergyResult()
+			self.results[key] = SynergyResult(self.sample_ids)
 		self.results[key][row,col,channel] = value
 	
 	def _add_metadata(self,**new_metadata):
@@ -446,42 +493,70 @@ class SynergyFile(list):
 			The separator character used in the file.
 		encoding : string specifying a supported encoding
 			The encoding of the file. This cannot be automatically detected.
+		verbose : boolean
+			whether to detail information about the parsing process. Mostly useful for debugging.
 	"""
-	def __init__(self,filename,separator="\t",encoding="iso-8859-1"):
+	def __init__(self,
+			filename,
+			separator="\t",
+			encoding="iso-8859-1",
+			verbose=False
+		):
 		super().__init__()
 		self.sep = separator
 		self._new_plate()
 		
+		self.verbose = verbose
+		
 		with open(filename,"r",encoding=encoding) as f:
 			self._line_buffer = LineBuffer(f.read().splitlines())
 		
 		self._parse_file()
 	
+	def report(self,message,newline=True):
+		if self.verbose:
+			print(
+					message,
+					end = "\n" if newline else "",
+				)
+	
 	def _parse_file(self):
 		while self._line_buffer:
 			for parser in (
 					self._parse_raw_data_matrix,
 					self._parse_raw_data_row,
+					wrap_variant(self._parse_raw_data_row,"blank"),
 					self._parse_raw_data_column,
+					wrap_variant(self._parse_raw_data_column,"blank"),
 					self._parse_results_matrix,
+					wrap_variant(self._parse_results_row,"well_id_and_conc"),
+					wrap_variant(self._parse_results_row,"well_id"),
 					self._parse_results_row,
+					wrap_variant(self._parse_results_column,"well_id_and_conc"),
+					wrap_variant(self._parse_results_column,"well_id"),
 					self._parse_results_column,
 					self._parse_single_matrix,
 					self._parse_single_row,
 					self._parse_single_column,
-					self._parse_procedure,
+					self._parse_other_metadata,
 					self._parse_gain_values,
 					self._parse_metadata,
+					self._parse_layout,
+					wrap_variant(self._parse_layout,"conc"),
+					self._parse_spectrum_column,
+					self._parse_blank_data,
 				):
 				try:
 					with self._line_buffer as line_iter:
 						parser(line_iter)
-				except FormatMismatch:
+				except FormatMismatch as e:
 					continue
 				except RepeatingData:
+					self.report("Found repeating data with {parser__name__}; created new plate.")
 					self._new_plate()
 					break
 				else:
+					self.report(f"Successful {parser.__name__:<22}")
 					break
 			else:
 				raise ValueError("File does not appear to have a valid or implemented format.")
@@ -507,15 +582,25 @@ class SynergyFile(list):
 		
 		self[-1]._add_metadata(**new_metadata)
 	
-	def _parse_procedure(self,line_iter):
-		format_assert( next(line_iter) == "Procedure Details" )
+	def _parse_other_metadata(self,line_iter):
+		header = next(line_iter)
+		try:
+			attribute = {
+				"Procedure Details": "procedure",
+				"Curve Fitting Results": "curve_fitting",
+				"DRCurve Fitting Results": "dr_curve_fitting",
+				"DRCurve Interpolations": "dr_curve_interpolations",
+			}[header]
+		except KeyError:
+			raise FormatMismatch
+		
 		format_assert( next(line_iter) == "" )
 		
-		procedure = []
+		content = []
 		for line in line_iter:
 			if line=="": break
-			procedure.append(line.replace(self.sep,"\t"))
-		self[-1]._add_metadata( procedure = "\n".join(procedure) )
+			content.append(line.replace(self.sep,"\t"))
+		self[-1]._add_metadata( **{attribute: "\n".join(content)} )
 	
 	def _parse_gain_values(self,line_iter):
 		format_assert( next(line_iter) == "Automatic gain values\t " )
@@ -564,19 +649,43 @@ class SynergyFile(list):
 			for col,number in zip(cols,numbers):
 				self[-1]._add_result(name,row,col,number)
 	
-	def _parse_results_row(self,line_iter):
+	def _parse_results_row(self,line_iter,variant=""):
 		format_assert( next(line_iter) == "Results" )
 		format_assert( next(line_iter) == "" )
 		
-		with ValueError_to_FormatMismatch():
-			Well,*names = next(line_iter).split(self.sep)
+		if variant=="well_id":
+			with ValueError_to_FormatMismatch():
+				Well_ID,Well,*names = next(line_iter).split(self.sep)
+			format_assert( Well_ID=="Well ID" )
+		elif variant=="well_id_and_conc":
+			with ValueError_to_FormatMismatch():
+				Well_ID,Well,Concdil,*names = next(line_iter).split(self.sep)
+			format_assert( Well_ID=="Well ID" )
+			format_assert( Concdil=="Conc/Dil" )
+		else:
+			with ValueError_to_FormatMismatch():
+				Well,*names = next(line_iter).split(self.sep)
 		format_assert( Well=="Well" )
 		
 		results = []
+		well_id = None
 		for line in line_iter:
 			if line=="": break
 			
-			well,*numbers = line.split(self.sep)
+			if variant=="well_id":
+				new_well_id,well,*numbers = line.split(self.sep)
+				well_id = new_well_id or well_id
+				format_assert( well_id == self[-1].layout[well] )
+			elif variant=="well_id_and_conc":
+				new_well_id,well,conc,*numbers = line.split(self.sep)
+				well_id = new_well_id or well_id
+				if conc:
+					format_assert( (well_id,parse_number(conc)) == self[-1].layout[well] )
+				else:
+					format_assert( well_id == self[-1].layout[well] )
+			else:
+				well,*numbers = line.split(self.sep)
+			
 			with ValueError_to_FormatMismatch():
 				row,col = split_well_name(well)
 			
@@ -590,23 +699,35 @@ class SynergyFile(list):
 		for name,row,col,number in results:
 			self[-1]._add_result(name,row,col,number)
 	
-	def _parse_results_column(self,line_iter):
+	def _parse_results_column(self,line_iter,variant=""):
 		format_assert( next(line_iter) == "Results" )
 		format_assert( next(line_iter) == "" )
 		
+		if variant.startswith("well_id"):
+			with ValueError_to_FormatMismatch():
+				Well_ID,*ids,last = next(line_iter).split(self.sep)
+			format_assert( Well_ID == "Well ID" )
+			format_assert( last == "" )
+		
 		with ValueError_to_FormatMismatch():
 			Well,*wells,last = next(line_iter).split(self.sep)
 			wells = [ split_well_name(well) for well in wells ]
 		format_assert( Well == "Well" )
 		format_assert( last == "" )
 		
+		if variant == "well_id_and_conc":
+			with ValueError_to_FormatMismatch():
+				concdil,*concs,last = next(line_iter).split(self.sep)
+			format_assert( concdil == "Conc/Dil" )
+			format_assert( last == "" )
+		
 		results = []
 		for line in line_iter:
 			if line=="": break
 			name,*numbers,last = line.split(self.sep)
+			format_assert( name != "Conc/Dil" )
 			format_assert( len(wells) == len(numbers) )
 			format_assert( last == "" )
-			
 			for attempt in TryFormats():
 				with attempt as format_parser:
 					numbers = [ format_parser(number) for number in numbers ]
@@ -617,35 +738,45 @@ class SynergyFile(list):
 			for (row,col),number in zip(wells,numbers):
 				self[-1]._add_result(name,row,col,number)
 	
-	def _parse_raw_data_column(self,line_iter):
+	def _parse_raw_data_column(self,line_iter,variant=""):
 		channel = next(line_iter)
 		format_assert( self.sep not in channel )
 		format_assert( next(line_iter) == "" )
 		
-		with ValueError_to_FormatMismatch():
-			Time,Temp,*wells = next(line_iter).split(self.sep)
+		if variant=="blank":
+			format_assert( channel.startswith("Blank ") )
+			with ValueError_to_FormatMismatch():
+				Time,*wells = next(line_iter).split(self.sep)
+		else:
+			with ValueError_to_FormatMismatch():
+				Time,Temp,*wells = next(line_iter).split(self.sep)
+			assert_temperature_label(Temp,channel)
 		format_assert( Time == "Time" )
-		assert_temperature_label(Temp,channel)
 		
 		results = []
 		for line in line_iter:
 			if line=="": break
 			with ValueError_to_FormatMismatch():
-				time,temperature,*numbers = line.split(self.sep)
+				if variant=="blank":
+					time,*numbers = line.split(self.sep)
+					temperature = np.nan
+				else:
+					time,temperature,*numbers = line.split(self.sep)
+					temperature = parse_number(temperature)
 				numbers = [ parse_number(number) for number in numbers ]
 				time = parse_time(time)
-				temperature = parse_number(temperature)
 			format_assert( len(numbers) == len(wells) )
 			results.append((time,temperature,numbers))
 		
 		for time,temperature,numbers in results:
 			if time==0 and all(np.isnan(numbers)):
 				continue
-			self[-1]._add_temperature(time,channel,temperature)
+			if variant!="blank":
+				self[-1]._add_temperature(time,channel,temperature)
 			for well,number in zip(wells,numbers):
 				self[-1]._add_raw_result(channel,*split_well_name(well),number,time)
 	
-	def _parse_raw_data_row(self,line_iter):
+	def _parse_raw_data_row(self,line_iter,variant=""):
 		channel = next(line_iter)
 		format_assert( self.sep not in channel )
 		format_assert( next(line_iter) == "" )
@@ -656,12 +787,17 @@ class SynergyFile(list):
 		format_assert( label == "Time" )
 		with ValueError_to_FormatMismatch():
 			times = [ parse_time(time) for time in times ]
-			label,*temperatures,last = next(line_iter).split(self.sep)
-		format_assert( last == "" )
-		assert_temperature_label(label,channel)
-		format_assert( len(temperatures) == len(times) )
-		with ValueError_to_FormatMismatch():
-			temperatures = [ float(temperature) for temperature in temperatures ]
+		
+		if variant == "blank":
+			temperatures = np.full_like(times,np.nan)
+		else:
+			with ValueError_to_FormatMismatch():
+				label,*temperatures,last = next(line_iter).split(self.sep)
+				temperatures = [ parse_number(temperature) for temperature in temperatures ]
+			format_assert( last == "" )
+				
+			assert_temperature_label(label,channel)
+			format_assert( len(temperatures) == len(times) )
 		
 		results = []
 		wells = []
@@ -670,6 +806,7 @@ class SynergyFile(list):
 			with ValueError_to_FormatMismatch():
 				well,*numbers,last = line.split(self.sep)
 				numbers = [ parse_number(number) for number in numbers ]
+				split_well_name(well)
 			format_assert( last == "" )
 			format_assert( len(numbers) == len(times) )
 			wells.append(well)
@@ -678,7 +815,8 @@ class SynergyFile(list):
 		for i,(time,temperature,*numbers) in enumerate(zip(times,temperatures,*results)):
 			if i>0 and time==0:
 				break
-			self[-1]._add_temperature(time,channel,temperature)
+			if variant != "blank":
+				self[-1]._add_temperature(time,channel,temperature)
 			for well,number in zip(wells,numbers):
 				self[-1]._add_raw_result(channel,*split_well_name(well),number,time)
 	
@@ -782,5 +920,82 @@ class SynergyFile(list):
 		
 		for (row,col),number in zip(wells,numbers):
 			self[-1]._add_result(channel,row,col,number)
-
+	
+	def _parse_layout(self,line_iter,variant=""):
+		format_assert( next(line_iter) == "Layout" )
+		
+		with ValueError_to_FormatMismatch():
+			empty,*cols = next(line_iter).split(self.sep)
+			cols = parse_cols(cols)
+		format_assert( empty == "" )
+		
+		layout = SynergyIndexable()
+		sample_ids = {}
+		expected_rows = row_iter()
+		
+		while True:
+			line = next(line_iter)
+			if line=="": break
+			
+			row  ,*labels ,well_id = line.split(self.sep)
+			format_assert( row == next(expected_rows) )
+			format_assert( empty == "" )
+			if variant=="conc":
+				empty,*condils,concdil = next(line_iter).split(self.sep)
+				format_assert( well_id == "Well ID" )
+				format_assert( concdil == "Conc/Dil" )
+				
+				with ValueError_to_FormatMismatch():
+					condils = [ parse_number(condil) for condil in condils ]
+				
+				ids = [
+						label if np.isnan(condil) else (label.partition(":")[0],condil)
+						for label,condil in zip(labels,condils)
+					]
+			else:
+				assert variant == ""
+				ids = labels
+			
+			for sample_id,col in zip(ids,cols):
+				layout[row,col] = sample_id
+				try:
+					sample_ids[sample_id].append((row,col))
+				except KeyError:
+					sample_ids[sample_id] = [(row,col)]
+		
+		self[-1].layout = layout
+		self[-1].sample_ids = sample_ids
+	
+	def _parse_blank_data(self,line_iter):
+		first_line = next(line_iter)
+		format_assert( "[Blank " in first_line )
+		format_assert( next(line_iter) == "" )
+		for line in line_iter:
+			if line=="": break
+		
+		warn("Data calculated from blanks will be ignored and not parsed.")
+	
+	def _parse_spectrum_column(self,line_iter):
+		channel,_,spectrum = next(line_iter).partition(":")
+		format_assert( spectrum == "Spectrum" )
+		format_assert( self.sep not in channel )
+		format_assert( next(line_iter) == "" )
+		
+		with ValueError_to_FormatMismatch():
+			Wavelength,*wells = next(line_iter).split(self.sep)
+			wells = [ split_well_name(well) for well in wells ]
+		format_assert( Wavelength == "Wavelength" )
+		
+		results = []
+		for line in line_iter:
+			if line=="": break
+			wavelength,*numbers = line.split(self.sep)
+			with ValueError_to_FormatMismatch():
+				wavelength = int(wavelength)
+				numbers = [ parse_number(number) for number in numbers ]
+			results.append((wavelength,numbers))
+		
+		for wavelength,numbers in results:
+			for (row,col),number in zip(wells,numbers):
+				self[-1]._add_result( (channel,wavelength), row, col, number )
 
